import tempfile
import sys
import contextlib
from io import StringIO


class External(object):
    # TODO: use real external
    def __init__(self, file):
        self._file = file

    def getvalue(self):
        return self.content

    @property
    def content(self):
        self._file.seek(0)
        return self._file.read()


@contextlib.contextmanager
def capture_stdout():
    orig = sys.stdout
    temp = tempfile.SpooledTemporaryFile(mode='w+t')

    with temp:
        sys.stdout = temp
        yield External(temp)
        sys.stdout = orig


@contextlib.contextmanager
def fake_stdin(text):
    stdin = sys.stdin
    sys.stdin = StringIO(text)
    yield
    sys.stdin = stdin


def record_calls(calls, function):
    def recorded(*args, **kwargs):
        calls.append((args, kwargs))
        return function(*args, **kwargs)
    return recorded


def _first_not_found(text, fragments):
    if not fragments:
        return None

    f = fragments[0]
    try:
        pos = text.index(f)
    except ValueError:
        return (text, f)

    return _first_not_found(text[pos + len(f):], fragments[1:])


assert not _first_not_found('abc', ('a', 'b', 'c'))
assert not _first_not_found('abc', ('a', 'c'))
assert not _first_not_found('abc', ('a', 'b'))
assert not _first_not_found('abc', ('b', 'c'))
assert ('c', 'a') == _first_not_found('abc', ('b', 'a'))
assert ('', 'a') == _first_not_found('abc', ('c', 'a'))
assert ('', 'b') == _first_not_found('abc', ('c', 'b', 'a'))

# or the equivalent
assert not _first_not_found('abc', 'abc')
assert not _first_not_found('abc', 'ac')
assert not _first_not_found('abc', 'ab')
assert not _first_not_found('abc', 'bc')
assert ('c', 'a') == _first_not_found('abc', 'ba')
assert ('', 'a') == _first_not_found('abc', 'ca')
assert ('', 'b') == _first_not_found('abc', 'cba')


class Assertions(object):

    def assertContainsInOrder(self, text, fragments):
        mismatch = _first_not_found(text, fragments)
        if mismatch:
            text_part, fragment = mismatch
            skipped = len(text) - len(text_part)
            msg = '\n'.join(
                (
                    '',
                    'Fragment',
                    ' >>> {} <<<'.format(fragment),
                    'not found in text' if not skipped
                    else 'not found in partial text ({} characters skipped)'
                    .format(skipped),
                    ' > ' + text_part.replace('\r', '').replace('\n', '\n > '),
                )
            )
            raise AssertionError(msg)



class REPO:
    '''Repo attributes'''
    TYPE = 'type'

    DIRECTORY = 'directory'
    VOLATILE = 'volatile'
    SERVE_INTERFACE = 'interface'
    SERVE_PORT = 'port'
    SERVE_USERNAME = 'username'
    SERVE_PASSWORD = 'password'

    # This is intentionally the same as SERVE_USERNAME and SERVE_PASSWORD
    # so that new users can get .pyrene configs with valid passwords
    USERNAME = SERVE_USERNAME
    PASSWORD = SERVE_PASSWORD
    DOWNLOAD_URL = 'download_url'
    UPLOAD_URL = 'upload_url'


class REPOTYPE:
    '''Values for REPO.TYPE'''
    DIRECTORY = 'directory'
    HTTP = 'http'



# Copyright 2013 Donald Stufft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Extracted from twine & modified by Krisztian Fekete
# Note: similar code exists in distutils.command.upload

from __future__ import absolute_import, division, print_function
from __future__ import unicode_literals

import hashlib
import os.path

import pkginfo
import pkg_resources
import requests
import re


class BDist(pkginfo.BDist):

    @property
    def py_version(self):
        pkgd = pkg_resources.Distribution.from_filename(self.filename)
        return pkgd.py_version


class SDist(pkginfo.SDist):

    py_version = None


def get_whl_py_version(filename):
    wheel_file_re = re.compile(
        r"""
        ^(?P<namever>(?P<name>.+?)(-(?P<ver>\d.+?))?)
        ((-(?P<build>\d.*?))?-(?P<pyver>.+?)-(?P<abi>.+?)-(?P<plat>.+?)
        \.whl|\.dist-info)$
        """,
        re.VERBOSE
    )

    wheel_info = wheel_file_re.match(filename)
    return wheel_info.group("pyver")


class Wheel(pkginfo.Wheel):

    @property
    def py_version(self):
        return get_whl_py_version(os.path.basename(self.filename))


DIST_TYPES = {
    "bdist_wheel": Wheel,
    "bdist_egg": BDist,
    "sdist": SDist,
}

DIST_EXTENSIONS = {
    ".whl": "bdist_wheel",
    ".egg": "bdist_egg",
    ".tar.bz2": "sdist",
    ".tar.gz": "sdist",
    ".zip": "sdist",
}


def get_dtype(filename):
    '''Determine distribution type'''
    for ext, dtype in DIST_EXTENSIONS.items():
        if filename.endswith(ext):
            return dtype

    raise ValueError(
        "Unknown distribution format: '%s'" %
        os.path.basename(filename)
    )


# "unit tests":
assert 'sdist' == get_dtype('pkg.tar.gz')
assert 'sdist' == get_dtype('pkg.zip')
# ? assert 'sdist' == get_dtype('pkg.tgz')
assert 'bdist_egg' == get_dtype('pkg.egg')
assert 'bdist_wheel' == get_dtype('pkg.whl')


def get_meta(filename, dtype):
    return DIST_TYPES[dtype](filename)


def upload(filename, signature, repository, username, password, comment):
    # Extract the metadata from the package
    dtype = get_dtype(filename)
    meta = get_meta(filename, dtype)

    # Fill in the data - send all the meta-data in case we need to
    # register a new release
    data = {
        # action
        ":action": "file_upload",
        "protcol_version": "1",

        # identify release
        "name": meta.name,
        "version": meta.version,

        # file content
        "filetype": dtype,
        "pyversion": meta.py_version,

        # additional meta-data
        "metadata_version": meta.metadata_version,
        "summary": meta.summary,
        "home_page": meta.home_page,
        "author": meta.author,
        "author_email": meta.author_email,
        "maintainer": meta.maintainer,
        "maintainer_email": meta.maintainer_email,
        "license": meta.license,
        "description": meta.description,
        "keywords": meta.keywords,
        "platform": meta.platforms,
        "classifiers": meta.classifiers,
        "download_url": meta.download_url,
        "supported_platform": meta.supported_platforms,
        "comment": comment,

        # PEP 314
        "provides": meta.provides,
        "requires": meta.requires,
        "obsoletes": meta.obsoletes,

        # Metadata 1.2
        "project_urls": meta.project_urls,
        "provides_dist": meta.provides_dist,
        "obsoletes_dist": meta.obsoletes_dist,
        "requires_dist": meta.requires_dist,
        "requires_external": meta.requires_external,
        "requires_python": meta.requires_python,
    }

    pypi_filename = os.path.basename(filename)

    with open(filename, "rb") as fp:
        content = fp.read()
        filedata = {
            "content": (pypi_filename, content),
        }
        if signature:
            filedata["gpg_signature"] = (pypi_filename + ".asc", signature)
        data["md5_digest"] = hashlib.md5(content).hexdigest()

    print("Uploading {0}".format(pypi_filename))

    session = requests.session()
    resp = session.post(
        repository,
        data=dict((k, v) for k, v in data.items() if v),
        files=filedata,
        auth=(username, password),
    )
    resp.raise_for_status()



# Py3 compatibility
from __future__ import print_function
from __future__ import unicode_literals

import abc
import os
import shutil
from .upload import upload
from .util import pip_install, PyPI, red, yellow
from .constants import REPO


class Repo(object):
    __metaclass__ = abc.ABCMeta

    ATTRIBUTES = (REPO.TYPE,)
    DEFAULTS = {}
    attributes = dict

    def __init__(self, name, attributes):
        super(Repo, self).__init__()
        self.name = name
        self.attributes = dict(attributes)

    def __getattr__(self, attribute):
        try:
            return self.attributes[attribute]
        except KeyError:
            try:
                return self.DEFAULTS[attribute]
            except KeyError:
                raise AttributeError(attribute)

    @abc.abstractmethod
    def get_as_pip_conf(self):
        pass

    @abc.abstractmethod
    def download_packages(self, package_spec, directory):
        pass

    @abc.abstractmethod
    def upload_packages(self, package_files):
        pass

    @abc.abstractmethod
    def serve(self):
        pass

    def print_attributes(self):
        def comment(text, color):
            return '# {}'.format(color(text))
        for attribute in self.ATTRIBUTES:
            if attribute in self.attributes:
                msg = (
                    '{} = {}'
                    .format(attribute, self.attributes[attribute])
                )
            elif attribute in self.DEFAULTS:
                msg = (
                    '# {} = {}'
                    .format(attribute, self.DEFAULTS[attribute])
                )
            else:
                msg = comment(attribute, yellow)

            print(msg)

        extra_attrs = sorted(
            set(self.attributes.keys()) - set(self.ATTRIBUTES)
        )
        if extra_attrs:
            print()
            print(comment('EXTRA ATTRIBUTES >>>', red))
            for attribute in extra_attrs:
                msg = '{} = {}'.format(attribute, self.attributes[attribute])
                print(msg)
            print(comment('<<< EXTRA ATTRIBUTES', red))


PIPCONF_BADREPO = '''\
[global]
# no package can be installed from a BadRepo
no-index = true
'''


class BadRepo(Repo):

    ATTRIBUTES = {}
    DEFAULTS = {}

    def get_as_pip_conf(self):
        return PIPCONF_BADREPO

    @property
    def printable_name(self):
        return red('{} (a misconfigured repo!)'.format(self.name))

    def download_packages(self, package_spec, directory):
        print(
            '{}: pretended to provide package "{}"'
            .format(self.printable_name, package_spec)
        )

    def upload_packages(self, package_files):
        if package_files:
            print(
                '{}: pretended to upload package files'
                .format(self.printable_name)
            )
            for file in package_files:
                print('  {}'.format(file))
        else:
            print('{}: nothing to upload'.format(self.printable_name))

    def serve(self):
        print('{}: is not served'.format(self.printable_name))


PIPCONF_DIRECTORYREPO = '''\
[global]
no-index = true
find-links = {directory}
'''


class DirectoryRepo(Repo):

    ATTRIBUTES = (
        REPO.TYPE,
        REPO.DIRECTORY,
        REPO.VOLATILE,
        REPO.SERVE_INTERFACE,
        REPO.SERVE_PORT,
        REPO.SERVE_USERNAME,
        REPO.SERVE_PASSWORD,
    )

    DEFAULTS = {
        REPO.SERVE_INTERFACE: '0.0.0.0',
        REPO.SERVE_PORT: '8080',
        REPO.VOLATILE: 'no',
    }

    def get_as_pip_conf(self):
        return PIPCONF_DIRECTORYREPO.format(directory=self.directory)

    def download_packages(self, package_spec, directory):
        pip_install(
            '--find-links', self.directory,
            '--no-index',
            '--download', directory.path,
            package_spec,
        )

    def ensure_repo_directory(self):
        if not os.path.isdir(self.directory):
            os.makedirs(self.directory)

    def upload_packages(self, package_files):
        self.ensure_repo_directory()

        destination = self.directory
        for source in package_files:
            shutil.copy2(source, destination)

    def serve(self, pypi_server=PyPI):
        self.ensure_repo_directory()

        server = pypi_server()
        server.directory = self.directory
        server.interface = getattr(self, REPO.SERVE_INTERFACE)
        server.port = getattr(self, REPO.SERVE_PORT)
        true = {'y', 'yes', 't', 'true'}
        server.volatile = getattr(self, REPO.VOLATILE).lower() in true

        try:
            username = getattr(self, REPO.SERVE_USERNAME)
            password = getattr(self, REPO.SERVE_PASSWORD)
        except AttributeError:
            pass
        else:
            server.add_user(username, password)

        server.serve()


PIPCONF_HTTPREPO = '''\
[global]
index-url = {download_url}
extra-index-url =
'''


class HttpRepo(Repo):

    ATTRIBUTES = (
        REPO.TYPE,
        REPO.UPLOAD_URL,
        REPO.DOWNLOAD_URL,
        REPO.USERNAME,
        REPO.PASSWORD,
    )

    DEFAULTS = {}

    def get_as_pip_conf(self):
        return PIPCONF_HTTPREPO.format(download_url=self.download_url)

    def download_packages(self, package_spec, directory):
        pip_install(
            '--index-url', self.download_url,
            '--download', directory.path,
            package_spec,
        )

    def upload_packages(self, package_files):
        for source in package_files:
            upload(
                source,
                signature=None,
                repository=self.upload_url,
                username=self.username,
                password=self.password,
                comment='Uploaded with Pyrene',
            )

    def serve(self):
        print('Externally served at url {}'.format(self.download_url))



# Py3 compatibility
from __future__ import print_function
from __future__ import unicode_literals

import os
from cmd import Cmd
import traceback
from .util import write_file, bold, red
from .network import Network, DirectoryRepo, UnknownRepoError
from .constants import REPO, REPOTYPE


class ShellError(Exception):

    def __init__(self, message):
        self.message = message


class BaseCmd(Cmd, object):

    def emptyline(self):
        pass

    def cmdloop(self, intro=None):
        while True:
            try:
                super(BaseCmd, self).cmdloop(intro)
                break
            except Exception:
                traceback.print_exc()
            except KeyboardInterrupt:
                print('^C')
            intro = ''

    def onecmd(self, line):
        try:
            return super(BaseCmd, self).onecmd(line)
        except ShellError as e:
            print(red('ERROR: {}'.format(e.message)))

    def do_EOF(self, line):
        '''
        Exit
        '''
        print('Bye!')
        return True
    do_bye = do_EOF

    def completenames(self, *args):
        # when there is only one completions, add an extra space
        completions = super(BaseCmd, self).completenames(*args)
        if len(completions) == 1:
            return [completions[0].rstrip() + ' ']
        return completions


REPO_ATTRIBUTE_COMPLETIONS = tuple(
    '{}='.format(a)
    for a in Network.REPO_ATTRIBUTES
)


class PyreneCmd(BaseCmd):

    intro = '''
    Pyrene provides tools to work with different repos of python packages.

    e.g. one might use three different repos in one project:

     - pypi.python.org       (globally shared)
     - private pypi instance (project/company specific,
                              pip needs to be configured to fetch from here)
     - developer cache       (~/.pip/local)

    For help on commands type {help} or {qmark}
    '''.format(help=bold('help'), qmark=bold('?'))

    @property
    def prompt(self):
        active_repo = self.network.active_repo
        prompt = (
            'Pyrene[{}]: '.format(active_repo) if active_repo
            else 'Pyrene: '
        )
        return prompt

    def __init__(self, network, directory):
        super(PyreneCmd, self).__init__()
        self.network = network
        self.__temp_dir = directory

    def precmd(self, line):
        self.network.reload()
        return super(PyreneCmd, self).precmd(line)

    def write_file(self, filename, content):
        write_file(filename, content)

    def abort_on_unknown_repository_name(self, repo_name, command):
        if repo_name not in self.network.repo_names:
            raise ShellError('Unknown repository: {}'.format(repo_name))

    def get_effective_repo_name(self, repo_name):
        return repo_name or self.network.active_repo

    def abort_on_missing_effective_repo_name(self, repo_name, command):
        if not self.get_effective_repo_name(repo_name):
            raise ShellError(
                (
                    'Command "{}" requires a repository,'
                    + ' but none was given or active'
                ).format(command)
            )

    def abort_on_nonexisting_effective_repo(self, repo_name, command):
        self.abort_on_missing_effective_repo_name(repo_name, command)
        self.abort_on_unknown_repository_name(
            self.get_effective_repo_name(repo_name),
            command
        )

    def abort_on_nonexisting_repo(self, repo_name, command):
        if not repo_name:
            raise ShellError(
                'Command "{}" requires a repository parameter'
                .format(command)
            )

        self.abort_on_unknown_repository_name(repo_name, command)

    def do_use(self, repo):
        self.abort_on_nonexisting_effective_repo(repo, 'use')

        repo = self.network.get_repo(repo)
        pip_conf = os.path.expanduser('~/.pip/pip.conf')
        self.write_file(pip_conf, repo.get_as_pip_conf().encode('utf8'))

    def help_use(self):
        help = '''
        Set up {pip} (when used outside of Pyrene) to use REPO by default.

        {warn}
        '''.format(
            pip=bold('pip'),
            warn=red('WARNING: Overwrites ~/.pip/pip.conf!')
        )
        print(help)

    def _get_destination_repo(self, word):
        if word.endswith(':'):
            repo_name = word[:-1]
            try:
                return self.network.get_repo(repo_name)
            except UnknownRepoError:
                raise ShellError(
                    'Repository {} is not known'.format(repo_name)
                )

        attributes = {'directory': word}
        return DirectoryRepo('Implicit({})'.format(word), attributes)

    def do_copy(self, line):
        '''
        Copy packages between repos

          copy SOURCE DESTINATION

        Where SOURCE can be either LOCAL-FILE or REPO:PACKAGE-SPEC
        DESTINATION can be either a REPO: or a directory.
        '''
        words = line.split()
        source, destination = words
        destination_repo = self._get_destination_repo(destination)
        local_file_source = ':' not in source

        if local_file_source:
            destination_repo.upload_packages([source])
        else:
            source_repo_name, _, package_spec = source.partition(':')
            try:
                source_repo = self.network.get_repo(source_repo_name)
            except UnknownRepoError:
                raise ShellError(
                    'Unknown repository {}'.format(source_repo_name)
                )

            # copy between repos with the help of temporary storage
            try:
                source_repo.download_packages(package_spec, self.__temp_dir)
                destination_repo.upload_packages(self.__temp_dir.files)
            finally:
                self.__temp_dir.clear()

    def do_work_on(self, repo):
        '''
        Make repo the active one.
        Commands working on a repo will use it as default for repo parameter.
        '''
        self.abort_on_nonexisting_repo(repo, 'work_on')
        self.network.active_repo = repo

    def __define_or_change_type(self, repo, repotype):
        effective_repo_name = self.get_effective_repo_name(repo)
        if effective_repo_name not in self.network.repo_names:
            self.network.define(effective_repo_name)
        self.network.active_repo = effective_repo_name
        self.network.set(effective_repo_name, REPO.TYPE, repotype)

    def do_http_repo(self, repo):
        '''
        [Re]define REPO as http package repository.

        http_repo REPO
        '''
        self.abort_on_missing_effective_repo_name(repo, 'http_repo')
        self.__define_or_change_type(repo, REPOTYPE.HTTP)

    def do_directory_repo(self, repo):
        '''
        [Re]define REPO as directory package repository.

        directory_repo REPO
        '''
        self.abort_on_missing_effective_repo_name(repo, 'directory_repo')
        self.__define_or_change_type(repo, REPOTYPE.DIRECTORY)

    def do_forget(self, repo):
        '''
        Drop definition of a repo.

        forget REPO
        '''
        self.abort_on_nonexisting_repo(repo, 'forget')
        self.network.forget(repo)

    def abort_on_invalid_active_repo(self, command):
        if self.network.active_repo not in self.network.repo_names:
            raise ShellError(
                'Command "{}" requires a valid repository to be worked on'
                .format(command)
            )

    def do_set(self, line):
        '''
        Set repository attributes on the active repo.

        set attribute=value

        # intended use:

        # directory repos:
        work_on developer-repo
        set type=directory
        set directory=package-directory

        # http repos:
        work_on company-private-repo
        set type=http
        set download-url=http://...
        set upload-url=http://...
        set username=user
        set password=pass
        '''
        self.abort_on_invalid_active_repo('set')
        repo = self.network.active_repo
        attribute, eq, value = line.partition('=')
        if not attribute:
            raise ShellError('command "set" requires a non-empty attribute')
        if not eq:
            raise ShellError('command "set" requires a value')
        self.network.set(repo, attribute, value)

    def complete_set(self, text, line, begidx, endidx):
        completions = set()
        complete_line = line[:endidx]
        words = complete_line.split()
        if '=' in words[-1]:
            attribute, _, value = words[-1].partition('=')
            if attribute == 'type':
                completions = set(Network.REPO_TYPES)
            if self.network.active_repo:
                attributes = self.network.get_attributes(
                    self.network.active_repo
                )
                if attribute in attributes:
                    completions.add(attributes[attribute])
        else:
            completions = REPO_ATTRIBUTE_COMPLETIONS
        return sorted(c for c in completions if c.startswith(text))

    def do_unset(self, attribute):
        '''
        Unset attribute on the active/default repo
        '''
        self.abort_on_invalid_active_repo('unset')
        if not attribute:
            raise ShellError('command "unset" requires a non-empty attribute')
        self.network.unset(self.network.active_repo, attribute)

    def complete_unset(self, text, line, begidx, endidx):
        repo_name = self.network.active_repo
        if not repo_name:
            return []
        repo = self.network.get_repo(repo_name)
        completions = repo.attributes.keys()
        return sorted(c for c in completions if c.startswith(text))

    def do_list(self, line):
        '''
        List known repos
        '''
        repo_names = self.network.repo_names
        print('Known repos:')
        print('    ' + '\n    '.join(repo_names))

    def do_show(self, repo):
        '''
        List repo attributes
        '''
        self.abort_on_nonexisting_effective_repo(repo, 'show')

        repo = self.network.get_repo(repo)
        repo.print_attributes()

    def do_setup_for_pypi_python_org(self, repo):
        '''
        Configure repo to point to the default package index
        https://pypi.python.org.
        '''
        effective_repo_name = self.get_effective_repo_name(repo)
        self.abort_on_nonexisting_repo(
            effective_repo_name, 'setup_for_pypi_python_org'
        )

        self.network.set(effective_repo_name, REPO.TYPE, REPOTYPE.HTTP)
        self.network.set(
            effective_repo_name,
            REPO.DOWNLOAD_URL,
            'https://pypi.python.org/simple/'
        )
        self.network.set(
            effective_repo_name,
            REPO.UPLOAD_URL,
            'https://pypi.python.org/pypi'
        )

    def do_setup_for_pip_local(self, repo):
        '''
        Configure repo to be directory based with directory `~/.pip/local`.
        Also makes that directory if needed.
        '''
        effective_repo_name = self.get_effective_repo_name(repo)
        self.abort_on_nonexisting_repo(
            effective_repo_name, 'setup_for_pip_local'
        )

        piplocal = os.path.expanduser('~/.pip/local')
        if not os.path.exists(piplocal):
            os.makedirs(piplocal)
        self.network.set(effective_repo_name, REPO.TYPE, REPOTYPE.DIRECTORY)
        self.network.set(effective_repo_name, REPO.DIRECTORY, piplocal)

    def do_serve(self, repo_name):
        '''
        Serve a local directory over http as a package index (like pypi).
        Intended for quick package exchanges.
        '''
        self.abort_on_nonexisting_effective_repo(repo_name, 'serve')

        repo = self.network.get_repo(repo_name)
        repo.serve()

    def complete_repo_name(self, text, line, begidx, endidx, suffix=''):
        return sorted(
            '{}{}'.format(name, suffix)
            for name in self.network.repo_names
            if name.startswith(text)
        )

    complete_http_repo = complete_repo_name
    complete_directory_repo = complete_repo_name
    complete_work_on = complete_repo_name
    complete_forget = complete_repo_name
    complete_show = complete_repo_name
    complete_use = complete_repo_name
    complete_setup_for_pypi_python_org = complete_repo_name
    complete_setup_for_pip_local = complete_repo_name
    complete_serve = complete_repo_name

    def complete_filenames(self, text, line, begidx, endidx):
        dir_prefix = '.'

        line_before = line[:begidx]
        if not line_before.endswith(' '):
            words = line_before.split()
            if len(words) > 1:
                dir_prefix = os.path.dirname(words[-1]) or '.'

        dir_prefix = os.path.abspath(dir_prefix)

        return sorted(
            (f + '/') if os.path.isdir(os.path.join(dir_prefix, f)) else f
            for f in os.listdir(dir_prefix)
            if f.startswith(text)
        )

    def complete_copy(self, text, line, begidx, endidx):
        line_before = line[:begidx]

        if line_before.endswith(':'):
            # no completion after "repo:"
            return []

        repos = []

        if line_before.endswith(' '):
            repos = self.complete_repo_name(
                text, line, begidx, endidx, suffix=':'
            )

        filenames = self.complete_filenames(text, line, begidx, endidx)
        return repos + filenames



# Py3 compatibility
from __future__ import print_function
from __future__ import unicode_literals

import pyrene.util as m
from .util import capture_stdout
import unittest

import os
import subprocess
import mock
from temp_dir import within_temp_dir

from tempfile import NamedTemporaryFile
from passlib.apache import HtpasswdFile


class Test_set_env(unittest.TestCase):

    def setUp(self):
        self.original_environ = os.environ.copy()

    def tearDown(self):
        os.environ.clear()
        os.environ.update(self.original_environ)

    def test_overwrite_existing_variable(self):
        os.environ['existing'] = '1'

        with m.set_env('existing', '2'):
            self.assertEqual('2', os.environ['existing'])

        self.assertEqual('1', os.environ['existing'])

    def test_make_new_variable(self):
        if 'new' in os.environ:
            del os.environ['new']

        with m.set_env('new', '2'):
            self.assertEqual('2', os.environ['new'])

        self.assertNotIn('new', os.environ)


TEST_SETUP_PY = b'''\
from distutils.core import setup

setup(
    name='foo',
    version='1.0',
    py_modules=['foo'],
)
'''


class Test_pip_install(unittest.TestCase):

    @within_temp_dir
    def test_copy_package(self):
        m.write_file('setup.py', TEST_SETUP_PY)
        m.write_file('foo.py', b'')
        subprocess.check_output(
            'python setup.py sdist'.split(),
            stderr=subprocess.STDOUT
        )
        os.mkdir('destination')

        with capture_stdout() as stdout:
            m.pip_install(
                '--download', 'destination',
                '--find-links', 'dist',
                '--no-index',
                'foo',
            )
            output = stdout.content

        # diagnostic - in case pip has failed
        print(output)

        self.assertTrue(os.path.exists('destination/foo-1.0.tar.gz'))


class Test_Directory(unittest.TestCase):

    @within_temp_dir
    def test_files(self):
        os.makedirs('a/directory')
        d = m.Directory('a')
        m.write_file('a/file1', b'')
        m.write_file('a/file2', b'')

        f1 = os.path.join('a', 'file1')
        f2 = os.path.join('a', 'file2')
        self.assertEqual([f1, f2], d.files)

    @within_temp_dir
    def test_clear(self):
        os.makedirs('a/directory')
        d = m.Directory('a')
        m.write_file('a/file1', b'')
        m.write_file('a/file2', b'')

        d.clear()

        self.assertEqual([], d.files)


class Test_generate_password(unittest.TestCase):

    def test_non_repeating(self):
        passwords = [m.generate_password() for _ in range(1000)]
        self.assertEqual(len(set(passwords)), len(passwords))


class Test_PyPI(unittest.TestCase):

    def setUp(self):
        self.server = m.PyPI()
        self.server.execute = mock.Mock()

    @property
    def executed_cmd(self):
        ARGS_LIST = 1
        return self.server.execute.mock_calls[0][ARGS_LIST][0]

    def test_serve_volatile(self):
        self.server.volatile = True
        self.server.serve()
        self.assertIn('--overwrite', self.executed_cmd)

    def test_serve_non_volatile(self):
        self.server.serve()
        self.assertNotIn('--overwrite', self.executed_cmd)

    def test_make_htpasswd(self):
        self.server.add_user('testuser', 'testpass')
        self.server.add_user('xtestuser', 'xtestpass')
        with NamedTemporaryFile() as file:
            self.server.make_htpasswd(file.name)

            ht = HtpasswdFile(file.name)
            self.assertEqual(['testuser', 'xtestuser'], sorted(ht.users()))
            self.assertTrue(ht.check_password('testuser', 'testpass'))

    def test_serve_with_credentials(self):
        self.server.add_user('u', 'p')
        self.server.serve()
        self.assertIn('--passwords', self.executed_cmd)

    def test_serve_no_user(self):
        self.server.serve()
        self.assertNotIn('--passwords', self.executed_cmd)



from __future__ import absolute_import, division, print_function
from __future__ import unicode_literals

import unittest
import pyrene.upload as m


class Test_get_whl_py_version(unittest.TestCase):

    def test_pip(self):
        self.assertEquals(
            'py2.py3',
            m.get_whl_py_version('pip-1.5.4-py2.py3-none-any.whl')
        )



# Py3 compatibility
from __future__ import print_function
from __future__ import unicode_literals

import os

import unittest
import mock
from temp_dir import within_temp_dir
import tempfile

import pyrene.shell as m
from pyrene.util import Directory
from pyrene.repos import Repo
from .util import capture_stdout, fake_stdin, Assertions, record_calls

write_file = m.write_file


class MockingNetwork(m.Network):
    '''get_repo returns BadRepo to prevent all destructive operations'''
    TYPE_TO_CLASS = {}

    def __init__(self, dot_pyrene, get_repo):
        super(MockingNetwork, self).__init__(dot_pyrene)
        self.__get_repo = get_repo

    def get_repo(self, repo_name):
        repo = self.__get_repo(repo_name)
        return (
            repo if repo
            else super(MockingNetwork, self).get_repo(repo_name)
        )


class Test_PyreneCmd_write_file(unittest.TestCase):

    def setUp(self):
        super(Test_PyreneCmd_write_file, self).setUp()
        self.network = mock.Mock(spec_set=m.Network)
        self.directory = mock.Mock(spec_set=Directory)
        self.cmd = m.PyreneCmd(
            network=self.network,
            directory=self.directory
        )

    @within_temp_dir
    def test_creates_subdirectories(self):
        self.cmd.write_file('subdir/test-file', b'somecontent')
        with open('subdir/test-file', 'rb') as f:
            self.assertEqual(b'somecontent', f.read())

    @within_temp_dir
    def test_does_not_resolve_tilde(self):
        self.cmd.write_file('~', b'somecontent')
        with open('~', 'rb') as f:
            self.assertEqual(b'somecontent', f.read())


class Test_PyreneCmd(Assertions, unittest.TestCase):

    def setUp(self):
        super(Test_PyreneCmd, self).setUp()
        self.repo1 = mock.Mock(spec_set=Repo)
        self.repo2 = mock.Mock(spec_set=Repo)
        self.somerepo = mock.Mock(spec_set=Repo)

        self.dot_pyrene = tempfile.NamedTemporaryFile()
        self.network = MockingNetwork(self.dot_pyrene.name, self.get_repo)

        # self.network = mock.Mock(spec_set=m.Network)
        # self.network.get_repo.configure_mock(side_effect=self.get_repo)
        self.directory = mock.Mock(spec_set=Directory)
        self.directory.files = ()
        self.cmd = m.PyreneCmd(
            network=self.network,
            directory=self.directory
        )
        self.cmd.write_file = mock.Mock()

    def tearDown(self):
        self.dot_pyrene.close()
        super(Test_PyreneCmd, self).tearDown()

    def define_repos(self, *repo_names):
        for repo_name in repo_names:
            self.network.define(repo_name)

    def define_repo(self, repo_name, attributes):
        self.network.define(repo_name)
        self.network.active_repo = repo_name
        for attr, value in attributes.items():
            self.network.set(repo_name, attr, value)

    def get_repo(self, repo_name):
        if repo_name == 'repo1':
            return self.repo1
        if repo_name == 'repo2':
            return self.repo2
        if repo_name == 'somerepo':
            return self.somerepo

    def test_use(self):
        self.network.define('somerepo')
        self.network.get_repo = mock.Mock(return_value=self.somerepo)
        pip_conf = 'someconf'
        self.somerepo.get_as_pip_conf.configure_mock(
            return_value=pip_conf
        )

        self.cmd.onecmd('use somerepo')

        self.cmd.write_file.assert_called_once_with(
            os.path.expanduser('~/.pip/pip.conf'),
            pip_conf.encode('utf8')
        )

    def test_use_with_unknown_repo(self):
        output = run_script(self.cmd, 'use undefined')

        self.assertContainsInOrder(
            output,
            ('ERROR:', 'Unknown repo', 'undefined')
        )
        self.assertEqual(0, self.cmd.write_file.call_count)

    def test_use_with_implicit_repo(self):
        output = run_script(
            self.cmd,
            '''
            directory_repo somerepo
            work_on somerepo
            use
            '''
        )

        self.assertNotIn('ERROR:', output)
        self.assertEqual(1, self.cmd.write_file.call_count)

    def test_use_with_missing_implicit_repo(self):
        output = run_script(self.cmd, 'use')

        self.assertIn('ERROR:', output)
        self.assertEqual(0, self.cmd.write_file.call_count)

    def test_copy_single_package(self):
        self.define_repos('repo1', 'repo2')
        self.directory.files = ['roman-2.0.0.zip']

        self.cmd.onecmd('copy repo1:roman==2.0.0 repo2:')

        self.repo1.download_packages.assert_called_once_with(
            'roman==2.0.0',
            self.directory
        )
        self.repo2.upload_packages.assert_called_once_with(['roman-2.0.0.zip'])

    def test_copy_uses_repo_to_download_packages(self):
        self.define_repos('repo1', 'repo2')

        self.cmd.onecmd('copy repo1:roman==2.0.0 repo2:')

        self.repo1.download_packages.assert_called_once_with(
            'roman==2.0.0',
            self.directory
        )

    def test_copy_uses_repo_to_upload_packages(self):
        self.define_repos('repo1', 'repo2')
        self.directory.files = ['roman-2.0.0.zip']

        self.cmd.onecmd('copy repo1:roman==2.0.0 repo2:')

        self.repo2.upload_packages.assert_called_once_with(['roman-2.0.0.zip'])

    def test_copy_package_with_dependencies(self):
        self.define_repos('repo1', 'repo2')
        package_files = ('pkg-1.0.0.tar.gz', 'dep-0.3.1.zip')
        self.directory.files = list(package_files)

        self.cmd.onecmd('copy repo1:pkg repo2:')

        self.repo1.download_packages.assert_called_once_with(
            'pkg',
            self.directory
        )
        self.repo2.upload_packages.assert_called_once_with(list(package_files))

    def test_copy_uploads_files(self):
        self.define_repos('somerepo')
        self.cmd.onecmd('copy /a/file somerepo:')

        self.somerepo.upload_packages.assert_called_once_with(
            ['/a/file']
        )

    def test_copy_from_unknown_repo(self):
        output = run_script(
            self.cmd,
            '''
            directory_repo somerepo
            copy unknown:pkg somerepo:
            '''
        )

        self.assertContainsInOrder(output, ('ERROR:', 'unknown'))
        self.assertEqual(0, self.somerepo.upload_packages.call_count)

    def test_copy_to_unknown_repo(self):
        self.directory.files = ['/tmp/downloaded-package-file']
        output = run_script(
            self.cmd,
            '''
            directory_repo somerepo
            copy somerepo:pkg unknown:
            '''
        )

        self.assertContainsInOrder(output, ('ERROR:', 'unknown'))
        self.assertEqual(0, self.somerepo.upload_packages.call_count)

    def test_get_destination_repo_on_repo1(self):
        self.define_repos('repo1')
        repo = self.cmd._get_destination_repo('repo1:')
        self.assertIs(self.repo1, repo)

    def test_get_destination_repo_on_directory(self):
        repo = self.cmd._get_destination_repo('/path/to/directory')
        self.assertEqual('/path/to/directory', repo.directory)

    def test_copy_with_directory_destination(self):
        self.define_repos('repo1')
        self.directory.files = ['a-pkg']
        self.cmd._get_destination_repo = mock.Mock(
            return_value=self.somerepo
        )

        self.cmd.onecmd('copy repo1:a /tmp/x')
        self.cmd._get_destination_repo.assert_called_once_with('/tmp/x')
        self.somerepo.upload_packages.assert_called_once_with(['a-pkg'])

    def test_copy_clears_directory_after_upload(self):
        self.define_repos('repo1', 'repo2')
        package_files = ('pkg-1.0.0.tar.gz', 'dep-0.3.1.zip')
        files = mock.PropertyMock(return_value=list(package_files))
        type(self.directory).files = files

        self.cmd.onecmd('copy repo1:pkg repo2:')

        files.assert_called_once_with()
        self.assertEqual(
            [mock.call.clear()],
            self.directory.mock_calls
        )

    def test_copy_clears_directory_even_if_download_fails(self):
        self.define_repos('repo1', 'repo2')
        self.repo1.download_packages.configure_mock(side_effect=Exception)

        try:
            self.cmd.onecmd('copy repo1:a repo2:')
        except:
            pass

        self.assertIn(mock.call.clear(), self.directory.mock_calls)

    def test_http_repo_defines_new_repo(self):
        output = run_script(
            self.cmd,
            '''
            http_repo new-repo
            list
            '''
        )

        self.assertIn('new-repo', output)

    def test_existing_repo_http_repo_changes_repo_type(self):
        run_script(
            self.cmd,
            '''
            directory_repo new-repo
            set attr=somevalue
            http_repo
            '''
        )

        self.assertEqual('new-repo', self.network.active_repo)
        repo = self.network.get_repo('new-repo')
        self.assertEqual(m.REPOTYPE.HTTP, repo.type)
        self.assertEqual('somevalue', repo.attr)

    def test_directory_repo_defines_new_repo(self):
        output = run_script(
            self.cmd,
            '''
            directory_repo new-repo
            list
            '''
        )

        self.assertIn('new-repo', output)

    def test_existing_repo_directory_repo_changes_repo_type(self):
        run_script(
            self.cmd,
            '''
            http_repo new-repo
            set attr=somevalue
            directory_repo
            '''
        )

        self.assertEqual('new-repo', self.network.active_repo)
        repo = self.network.get_repo('new-repo')
        self.assertEqual(m.REPOTYPE.DIRECTORY, repo.type)
        self.assertEqual('somevalue', repo.attr)

    def test_forget(self):
        self.network.define('somerepo')

        self.cmd.onecmd('forget somerepo')

        self.assertNotIn('somerepo', self.network.repo_names)

    def test_unset(self):
        output = run_script(
            self.cmd,
            '''
            http_repo repo
            set someattribute=value
            unset someattribute
            set someotherattribute=othervalue
            show repo
            '''
        )

        self.assertNotIn('someattribute', output)
        self.assertIn('someotherattribute', output)

    def test_list(self):
        self.define_repos('S1', '#@!')

        with capture_stdout() as stdout:
            self.cmd.onecmd('list')
            output = stdout.content

        self.assertIn('S1', output)
        self.assertIn('#@!', output)

    def test_show(self):
        output = run_script(
            self.cmd,
            '''
            directory_repo repo
            set name=SHRP1
            set type=??
            show repo
            '''
        )

        self.assertIn('SHRP1', output)
        self.assertIn('??', output)

    def test_complete_set_on_attribute(self):
        completion = self.cmd.complete_set('', 'set atibute=value', 4, 4)
        self.assertEqual(set(m.REPO_ATTRIBUTE_COMPLETIONS), set(completion))

    def test_complete_set_on_attribute_ty(self):
        completion = self.cmd.complete_set('ty', 'set ty=value', 4, 6)
        self.assertEqual({'type='}, set(completion))

    def test_complete_set_on_type_value_fi(self):
        completion = self.cmd.complete_set('di', 'set type=di', 9, 11)
        self.assertEqual(['directory'], completion)

    def test_complete_set_on_empty_type_value(self):
        completion = self.cmd.complete_set('', 'set type=', 9, 9)
        self.assertEqual(set(m.Network.REPO_TYPES), set(completion))

    def test_complete_set_completes_existing_attribute_value(self):
        run_script(
            self.cmd,
            '''
            directory_repo repo
            set attr=somevalue
            '''
        )
        completion = self.cmd.complete_set('', 'set attr=', 9, 9)
        self.assertEqual({'somevalue'}, set(completion))

    def test_complete_set_on_value(self):
        completion = self.cmd.complete_set(
            'attribute=', 'set re attribute=value', 7, 17
        )
        self.assertEqual(set(), set(completion))

    def test_complete_unset_on_attribute_name(self):
        self.define_repo('repo', {'a': '1', 'b': '2'})
        completion = self.cmd.complete_unset('', 'unset ', 6, 6)
        self.assertEqual({'a', 'b'}, set(completion))

    def test_complete_repo_name(self):
        self.define_repos('repo', 'repo2')
        completion = self.cmd.complete_repo_name('', 'cmd ', 4, 4)
        self.assertEqual({'repo', 'repo2'}, set(completion))

    def test_complete_repo_name_with_suffix(self):
        self.define_repos('repo', 'repo2')
        completion = self.cmd.complete_repo_name('', 'cmd ', 4, 4, suffix=':')
        self.assertEqual({'repo:', 'repo2:'}, set(completion))

    def test_complete_repo_name_returns_sorted_output(self):
        self.define_repos('c-repo', 'repo-b', 'repo-a')
        completion = self.cmd.complete_repo_name('re', 'cmd re', 4, 6)
        self.assertEqual(['repo-a', 'repo-b'], completion)

    def test_complete_copy_completes_repos(self):
        self.define_repos('repo', 'repo2')
        completion = self.cmd.complete_copy('', 'copy ', 5, 5)
        self.assertTrue({'repo:', 'repo2:'}.issubset(set(completion)))

    @within_temp_dir
    def test_complete_filenames(self):
        os.makedirs('a')
        os.makedirs('c')
        write_file('ef', b'')

        completion = self.cmd.complete_filenames('', 'cmd ', 4, 4)

        self.assertEqual(['a/', 'c/', 'ef'], completion)

    @within_temp_dir
    def test_complete_filenames_with_prefix(self):
        os.makedirs('aa')

        completion = self.cmd.complete_filenames('a', 'cmd a', 4, 5)

        self.assertEqual(['aa/'], completion)

    @within_temp_dir
    def test_complete_filenames_with_nonexistent_prefix(self):
        completion = self.cmd.complete_filenames('a', 'cmd a/b/', 4, 8)

        self.assertEqual([], completion)

    @within_temp_dir
    def test_complete_filenames_with_subpaths(self):
        write_file('a/b/c', b'')

        completion = self.cmd.complete_filenames('', 'cmd a/b/', 6, 6)

        self.assertEqual(['b/'], completion)

    @within_temp_dir
    def test_complete_copy_completes_directories(self):
        os.makedirs('dir3/dir2/dir1')

        completion = self.cmd.complete_copy('', 'copy ', 4, 4)

        self.assertTrue({'dir3/'}.issubset(set(completion)))

    @within_temp_dir
    def test_complete_copy_does_not_complete_repos_after_slash(self):
        os.makedirs('dir')
        self.define_repos('repo', 'repo2')

        completion = self.cmd.complete_copy('', 'copy ./', 7, 7)

        self.assertEqual(['dir/'], completion)

    @within_temp_dir
    def test_complete_copy_does_not_complete_filenames_after_a_repo(self):
        os.makedirs('dir')
        self.define_repos('repo', 'repo2')

        completion = self.cmd.complete_copy('', 'copy repo:', 10, 10)

        self.assertEqual([], completion)

    def test_setup_for_pypi_python_org(self):
        self.define_repos('repo')

        self.cmd.onecmd('setup_for_pypi_python_org repo')

        repo = self.network.get_repo('repo')
        self.assertIn('pypi.python.org', repo.download_url)

    def test_setup_for_pip_local(self):
        self.define_repos('repo')

        self.cmd.onecmd('setup_for_pip_local repo')

        repo = self.network.get_repo('repo')
        self.assertIn('pip/local', repo.directory)

    def test_serve(self):
        self.define_repos('repo1')

        self.cmd.onecmd('serve repo1')

        self.repo1.serve.assert_called_once_with()

    def test_network_reload_called_before_every_command_in_the_loop(self):
        calls = []
        self.network.reload = record_calls(calls, self.network.reload)
        self.define_repos('repo-a')

        run_script(self.cmd, '\nlist\n')

        # expect 3 calls, one for each of
        # empty line
        # list
        # EOF
        self.assertEqual(3, len(calls))

    def test_define_sets_default_repo(self):
        script = '''
        http_repo repo
        set someattr=somevalue
        show repo
        '''
        output = run_script(self.cmd, script)
        self.assertContainsInOrder(output, ['someattr', 'somevalue'])

    def test_work_on_sets_prompt(self):
        output = run_script(
            self.cmd,
            '''
            http_repo somerepo
            work_on somerepo
            '''
        )
        self.assertIn('Pyrene[somerepo]: ', output)


def run_script(cmd, script):
    with fake_stdin(script):
        with capture_stdout() as stdout:
            cmd.cmdloop()
            return stdout.content


#########################################################
import nose.util


# nose specific: uses test generators for checking all commands
class Test_PyreneCmd_repo_parameter_checking(Assertions):

    def setup(self):
        self.dot_pyrene = tempfile.NamedTemporaryFile()
        self.network = m.Network(self.dot_pyrene.name)

        self.directory = mock.Mock(spec_set=Directory)
        self.directory.files = ('dummy')
        self.cmd = m.PyreneCmd(
            network=self.network,
            directory=self.directory
        )
        self.cmd.write_file = mock.Mock()

    def teardown(self):
        self.dot_pyrene.close()

    def test_requires_explicit_repo_parameter(self):
        commands = [
            'forget',
            'work_on',
        ]
        for command in commands:
            yield self.check_requires_repo_parameter, command

    def check_requires_repo_parameter(self, command):
        output = run_script(
            self.cmd,
            '''
            directory_repo somerepo
            work_on somerepo
            {}
            '''.format(command)
        )
        self.assertContainsInOrder(output, ('ERROR', command, 'requires'))

    def test_active_repo_works(self):
        commands = [
            'use',
            # 'forget',
            'show',
            'setup_for_pip_local',
            'setup_for_pypi_python_org',
            'serve',
            # 'work_on',
            'directory_repo',
            'http_repo',
        ]
        for command in commands:
            yield self.check_active_repo_works_for, command

    def check_active_repo_works_for(self, command):
        output = run_script(
            self.cmd,
            '''
            http_repo somerepo
            set download_url=http://example.com:8080/simple
            work_on somerepo
            {}
            '''.format(command)
        )
        nose.tools.assert_not_in('ERROR', output)

    def test_missing_active_repo_error_message(self):
        commands = [
            'use',
            'forget',
            'show',
            'setup_for_pip_local',
            'setup_for_pypi_python_org',
            'serve',
            'work_on',
            'directory_repo',
            'http_repo',
        ]
        for command in commands:
            yield self.check_missing_active_repo_error_message, command

    def check_missing_active_repo_error_message(self, command):
        output = run_script(
            self.cmd,
            '''
            {}
            '''.format(command)
        )
        self.assertContainsInOrder(output, ('ERROR', command, 'requires'))

    def test_active_repo_only_error_message(self):
        commands = [
            ('set', 'attr=value'),
            ('unset', 'attr'),
        ]
        for command, param in commands:
            yield self.check_active_repo_only_error_message, command, param

    def check_active_repo_only_error_message(self, command, param):
        output = run_script(
            self.cmd,
            '''
            {} {}
            '''.format(command, param)
        )
        self.assertContainsInOrder(
            output,
            ('ERROR', command, 'requires', 'work')
        )

    def test_unknown_repo_error_message(self):
        commands = [
            'use',
            'forget',
            'show',
            'setup_for_pip_local',
            'setup_for_pypi_python_org',
            'serve',
            'work_on',
            # 'directory_repo',
            # 'http_repo',
        ]
        for command in commands:
            yield self.check_unknown_repo_error_message, command

    def check_unknown_repo_error_message(self, command):
        output = run_script(
            self.cmd,
            '''
            {} undefined-repo
            '''.format(command)
        )
        self.assertContainsInOrder(output, ('ERROR', 'undefined-repo'))

    def test_set_without_parameter(self):
        output = run_script(
            self.cmd,
            '''
            directory_repo repo
            set
            '''
        )
        self.assertContainsInOrder(
            output,
            ('ERROR', 'set', 'requires', 'attribute')
        )

    def test_set_without_equal_sign(self):
        output = run_script(
            self.cmd,
            '''
            directory_repo repo
            set attr
            '''
        )
        self.assertContainsInOrder(
            output,
            ('ERROR', 'set', 'requires', 'value')
        )

    def test_unset_without_attribute(self):
        output = run_script(
            self.cmd,
            '''
            directory_repo repo
            unset
            '''
        )
        self.assertContainsInOrder(
            output,
            ('ERROR', 'unset', 'requires', 'attribute')
        )

# TODO: tests for error cases in copy



# Py3 compatibility
from __future__ import print_function
from __future__ import unicode_literals

import tempfile
import os
import sys
import shutil
from .network import Network
from .util import Directory
from .shell import PyreneCmd


def main():
    tempdir = tempfile.mkdtemp(suffix='.pyrene')
    cmd = PyreneCmd(
        Network(os.path.expanduser('~/.pyrene')),
        Directory(tempdir),
    )
    line = ' '.join(sys.argv[1:])
    try:
        if line:
            cmd.onecmd(line)
        else:
            cmd.cmdloop()
    finally:
        shutil.rmtree(tempdir)


if __name__ == '__main__':
    main()



# Py3 compatibility
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
import subprocess
import signal
import contextlib
from tempfile import NamedTemporaryFile
from passlib.apache import HtpasswdFile
import termcolor


def red(text):
    return termcolor.colored(text, 'red')


def green(text):
    return termcolor.colored(text, 'green')


def yellow(text):
    return termcolor.colored(text, 'yellow')


def bold(text):
    return termcolor.colored(text, attrs={'bold'})


@contextlib.contextmanager
def set_env(variable, value):
    exists = variable in os.environ
    original_value = os.environ.get(variable, '')
    os.environ[variable] = value
    yield
    os.environ[variable] = original_value
    if not exists:
        del os.environ[variable]


def pip_install(*args):
    '''
    Run pip install ...

    Explicitly ignores user's config.
    '''
    pip_cmd = os.path.join(os.path.dirname(sys.executable), 'pip')
    with set_env('PIP_CONFIG_FILE', os.devnull):
        cmd = [pip_cmd, 'install'] + list(args)
        print(' '.join(cmd))
        subprocess.call(cmd, stdout=sys.stdout, stderr=sys.stderr)


class PyPI(object):

    def __init__(self):
        self.directory = '.'
        self.volatile = False
        self.interface = '0.0.0.0'
        self.port = '8080'
        self.users = {}

    def add_user(self, username, password):
        self.users[username] = password

    def make_htpasswd(self, filename):
        ht = HtpasswdFile(path=filename, new=True)
        for username, password in self.users.items():
            ht.set_password(username, password)
        ht.save()

    def serve(self):
        pypi_srv = os.path.join(os.path.dirname(sys.executable), 'pypi-server')
        with NamedTemporaryFile() as password_file:
            self.make_htpasswd(password_file.name)

            cmd = [
                pypi_srv,
                '--interface', self.interface,
                '--port', self.port,
            ] + (
                ['--passwords', password_file.name] if self.users else []
            ) + [
                '--disable-fallback',
            ] + (
                ['--overwrite'] if self.volatile else []
            ) + [self.directory]

            self.execute(cmd)

    def execute(self, cmd):
        process = subprocess.Popen(cmd)
        try:
            process.wait()
        except KeyboardInterrupt:
            os.kill(process.pid, signal.SIGHUP)
            process.wait()
        print()


def write_file(path, content):
    try:
        os.makedirs(os.path.dirname(path))
    except OSError:
        pass
    with open(path, 'wb') as file:
        file.write(content)


class Directory(object):

    def __init__(self, path):
        self.path = os.path.normpath(path)

    @property
    def files(self):
        candidates = (
            os.path.join(self.path, f) for f in os.listdir(self.path)
        )
        return sorted(f for f in candidates if os.path.isfile(f))

    def clear(self):
        for path in self.files:
            os.remove(path)


def generate_password():
    import binascii
    return binascii.hexlify(os.urandom(10))



# Py3 compatibility
from __future__ import print_function
from __future__ import unicode_literals

import pyrene.network as m
import unittest

import os
import tempfile
from temp_dir import within_temp_dir
from pyrene.util import write_file
from pyrene.constants import REPO, REPOTYPE
from pyrene.repos import Repo


class Test_Network_create(unittest.TestCase):

    @within_temp_dir
    def test_missing_repo_store_define_creates_it(self):
        network = m.Network('repo_store')
        network.define('repo')

        self.assertTrue(os.path.isfile('repo_store'))


class Test_Network(unittest.TestCase):

    def setUp(self):
        fd, self.repo_store = tempfile.mkstemp()
        os.close(fd)
        self.network = m.Network(self.repo_store)

    def tearDown(self):
        os.remove(self.repo_store)

    def test_get_repo_fails_on_undefined_repo(self):
        with self.assertRaises(m.UnknownRepoError):
            self.network.get_repo('undefined')

    def test_get_repo_returns_badrepo_on_missing_repo_type(self):
        self.network.define('no-type')
        self.network.set('no-type', 'attr', 'attr-value')
        repo = self.network.get_repo('no-type')
        self.assertIsInstance(repo, m.BadRepo)
        self.assertEqual('attr-value', repo.attr)

    def test_get_repo_returns_repo(self):
        self.network.define('repo')
        self.network.set('repo', 'type', 'http')
        repo = self.network.get_repo('repo')
        self.assertIsInstance(repo, Repo)

    def test_get_repo_fails_on_unknown_repo_type(self):
        self.network.define('repo')
        self.network.set('repo', 'type', 'unknown!')
        repo = self.network.get_repo('repo')
        self.assertIsInstance(repo, m.BadRepo)

    def test_get_repo_sets_repo_name(self):
        self.network.define('r!')
        repo = self.network.get_repo('r!')
        self.assertEqual('r!', repo.name)

    def test_get_repo_with_empty_repo_name_returns_active_repo(self):
        self.network.define('activerepo')
        self.network.set('activerepo', 'type', 'http')
        self.network.active_repo = 'activerepo'

        repo = self.network.get_repo('')

        self.assertEqual('activerepo', repo.name)

    def make_file_repo(self, directory):
        self.network.define('repo')
        self.network.set('repo', REPO.TYPE, REPOTYPE.DIRECTORY)
        self.network.set('repo', REPO.DIRECTORY, directory)

    def test_directory_is_available_on_file_repo(self):
        self.make_file_repo('/a/repo/dir')

        repo = self.network.get_repo('repo')
        self.assertEqual('/a/repo/dir', repo.directory)

    def test_set_is_persistent(self):
        self.make_file_repo('/a/repo/dir')

        other_network = m.Network(self.repo_store)
        repo = other_network.get_repo('repo')
        self.assertEqual('/a/repo/dir', repo.directory)

    def test_unset_is_persistent(self):
        self.make_file_repo('/a/repo/dir')
        self.network.unset('repo', REPO.DIRECTORY)

        other_network = m.Network(self.repo_store)
        repo = other_network.get_repo('repo')
        with self.assertRaises(AttributeError):
            repo.directory
        # self.assertEqual('/a/repo/dir', repo.directory)

    def test_forget_is_persistent(self):
        self.make_file_repo('/a/repo/dir')
        self.network.forget('repo')

        other_network = m.Network(self.repo_store)
        with self.assertRaises(m.UnknownRepoError):
            other_network.get_repo('repo')

    def test_repo_names(self):
        self.network.define('r1')
        self.network.define('r4')
        self.assertEqual({'r1', 'r4'}, set(self.network.repo_names))

    def test_get_attributes(self):
        self.network.define('r1')
        self.network.set('r1', 'ame', '2')
        self.network.set('r1', 'nme', '22')
        self.network.define('r2')
        self.network.set('r2', 'name', 'fixed')

        self.assertDictEqual(
            {'ame': '2', 'nme': '22'},
            self.network.get_attributes('r1')
        )
        self.assertDictEqual(
            {'name': 'fixed'},
            self.network.get_attributes('r2')
        )

    def test_get_attributes_on_undefined_repo(self):
        with self.assertRaises(m.UnknownRepoError):
            self.network.get_attributes('undefined-repo')


TEST_CONFIG = b'''\
[repo:1]
type=directory
directory=/tmp
'''


class Test_Network_dot_pyrene(unittest.TestCase):

    def setUp(self):
        fd, self.repo_store = tempfile.mkstemp()
        os.close(fd)
        self.network = m.Network(self.repo_store)

    def tearDown(self):
        os.remove(self.repo_store)

    def test_reload(self):
        self.network.define('repo')

        write_file(self.repo_store, TEST_CONFIG)
        self.network.reload()

        self.assertEqual(['1'], self.network.repo_names)



# Py3 compatibility
from __future__ import print_function
from __future__ import unicode_literals

import os
try:
    from ConfigParser import RawConfigParser
except ImportError:
    def RawConfigParser():
        import configparser
        return configparser.ConfigParser(interpolation=None)

from .repos import BadRepo, DirectoryRepo, HttpRepo
from .constants import REPO, REPOTYPE


class UnknownRepoError(NameError):
    '''Repo is not defined at all'''


class Network(object):

    REPO_TYPES = {
        REPOTYPE.DIRECTORY,
        REPOTYPE.HTTP,
    }

    REPO_ATTRIBUTES = set(
        DirectoryRepo.ATTRIBUTES
    ).union(
        set(HttpRepo.ATTRIBUTES)
    )

    REPO_SECTION_PREFIX = 'repo:'

    TYPE_TO_CLASS = {
        REPOTYPE.DIRECTORY: DirectoryRepo,
        REPOTYPE.HTTP: HttpRepo,
    }

    # name of active/default/ repo
    active_repo = None

    def __init__(self, filename):
        self._repo_store_filename = filename
        self._config = None
        self.active_repo = None
        self.reload()

    def reload(self):
        self._config = RawConfigParser()
        if os.path.exists(self._repo_store_filename):
            self._config.read(self._repo_store_filename)

    def _save(self):
        with open(self._repo_store_filename, 'wt') as f:
            self._config.write(f)

    def get_repo(self, repo_name):
        repo_name = repo_name or self.active_repo
        repokey = self.REPO_SECTION_PREFIX + repo_name
        if not self._config.has_section(repokey):
            raise UnknownRepoError(repo_name)

        attributes = self.get_attributes(repo_name)
        repo_type = attributes.get(REPO.TYPE)

        repo_class = self.TYPE_TO_CLASS.get(repo_type, BadRepo)
        return repo_class(repo_name, attributes)

    def define(self, repo_name):
        repokey = self.REPO_SECTION_PREFIX + repo_name
        self._config.add_section(repokey)
        self._save()

    def forget(self, repo_name):
        repokey = self.REPO_SECTION_PREFIX + repo_name
        self._config.remove_section(repokey)
        self._save()

    def set(self, repo_name, attribute, value):
        repokey = self.REPO_SECTION_PREFIX + repo_name
        self._config.set(repokey, attribute, value)
        self._save()

    def unset(self, repo_name, attribute):
        repokey = self.REPO_SECTION_PREFIX + repo_name
        self._config.remove_option(repokey, attribute)
        self._save()

    @property
    def repo_names(self):
        return [
            section[len(self.REPO_SECTION_PREFIX):]
            for section in self._config.sections()
            if section.startswith(self.REPO_SECTION_PREFIX)
        ]

    def get_attributes(self, repo_name):
        repokey = self.REPO_SECTION_PREFIX + repo_name
        if not self._config.has_section(repokey):
            raise UnknownRepoError(repo_name)

        attributes = {
            attribute: self._config.get(repokey, attribute)
            for attribute in self._config.options(repokey)
        }
        return attributes



# Py3 compatibility
from __future__ import print_function
from __future__ import unicode_literals

import unittest
import mock
import os
from temp_dir import within_temp_dir

import pyrene.repos as m
from pyrene.constants import REPO, REPOTYPE
from .util import capture_stdout, Assertions


class Test_BadRepo(unittest.TestCase):

    def setUp(self):
        self.repo = m.BadRepo('repo', {})

    def test_download_package(self):
        self.repo.download_packages('a', '.')

    def test_upload_packages(self):
        self.repo.upload_packages(['a'])


class Test_DirectoryRepo(Assertions, unittest.TestCase):

    def make_repo(self, attrs, name='repo'):
        attrs = attrs.copy()
        attrs[REPO.TYPE] = REPOTYPE.DIRECTORY
        return m.DirectoryRepo(name, attrs)

    def test_attributes(self):
        repo = self.make_repo({'directory': 'dir@', 'type': 'directory'})
        self.assertEqual('directory', repo.type)
        self.assertEqual('dir@', repo.directory)

    def test_incomplete_repo_get_as_pip_conf(self):
        repo = self.make_repo({})
        with self.assertRaises(AttributeError):
            repo.get_as_pip_conf()

    def test_get_as_pip_conf(self):
        directory = '/path/to/repo'
        repo = self.make_repo({REPO.DIRECTORY: directory})
        self.assertIn(directory, repo.get_as_pip_conf())

    def test_serve_without_upload_user(self):
        repo = self.make_repo({REPO.DIRECTORY: '.'})
        pypi = mock.Mock()
        repo.serve(pypi)

    def test_serve_with_upload_user(self):
        attrs = {
            REPO.DIRECTORY: '.',
            REPO.SERVE_USERNAME: 'tu',
            REPO.SERVE_PASSWORD: 'tp',
        }
        repo = self.make_repo(attrs)
        pypi = mock.Mock()
        repo.serve(pypi)

    @within_temp_dir
    def test_serve_creates_nonexistent_repo_directory(self):
        repo = self.make_repo({REPO.DIRECTORY: 'missing'})
        pypi = mock.Mock()

        repo.serve(pypi)

        self.assertTrue(os.path.isdir('missing'))

    @within_temp_dir
    def test_upload_packages_creates_nonexistent_repo_directory(self):
        repo = self.make_repo({REPO.DIRECTORY: 'missing'})

        repo.upload_packages([])

        self.assertTrue(os.path.isdir('missing'))

    def test_print_attributes(self):
        with capture_stdout() as stdout:
            self.make_repo({}).print_attributes()
            output = stdout.content

        self.assertContainsInOrder(output, m.DirectoryRepo.ATTRIBUTES)
        self.assertNotIn(REPO.DOWNLOAD_URL, output)


class Test_HttpRepo(Assertions, unittest.TestCase):

    def make_repo(self, attrs, name='repo'):
        attrs = attrs.copy()
        attrs[REPO.TYPE] = REPOTYPE.HTTP
        return m.HttpRepo(name, attrs)

    def test_attributes(self):
        repo = self.make_repo(
            {REPO.DOWNLOAD_URL: 'https://priv.repos.org/simple'}
        )
        self.assertEqual('http', repo.type)
        self.assertEqual('https://priv.repos.org/simple', repo.download_url)

    def test_serve(self):
        repo = self.make_repo(
            {REPO.DOWNLOAD_URL: 'https://priv.repos.org/simple'}
        )

        with capture_stdout() as stdout:
            repo.serve()
            output = stdout.content

        self.assertIn('https://priv.repos.org/simple', output)

    def test_incomplete_repo_get_as_pip_conf(self):
        repo = self.make_repo({})
        with self.assertRaises(AttributeError):
            repo.get_as_pip_conf()

    def test_get_as_pip_conf(self):
        url = 'http://download/url'
        repo = self.make_repo({REPO.DOWNLOAD_URL: url})
        self.assertIn(url, repo.get_as_pip_conf())

    def test_print_attributes(self):
        with capture_stdout() as stdout:
            self.make_repo({}).print_attributes()
            output = stdout.content

        self.assertContainsInOrder(output, m.HttpRepo.ATTRIBUTES)
        self.assertNotIn(REPO.DIRECTORY, output)
